@@ -54,6 +54,7 @@ from guppylang_internals.tys.ty import (
 
 if TYPE_CHECKING:
     from guppylang_internals.definition.struct import StructField
+    from guppylang_internals.tys.parsing import TypeParsingCtx
 
 
 #: A "place" is a description for a storage location of a local value that users
@@ -116,6 +117,10 @@ class Variable:
         """Returns a new `Variable` instance with an updated definition location."""
         return replace(self, defined_at=node)
 
+    def add_flags(self, flags: InputFlags) -> "Variable":
+        """Returns a new `Variable` instance with updated flags."""
+        return replace(self, flags=self.flags | flags)
+
 
 @dataclass(frozen=True, kw_only=True)
 class ComptimeVariable(Variable):
@@ -507,6 +512,13 @@ class Context(NamedTuple):
     locals: Locals[str, Variable]
     generic_params: dict[str, Parameter]
 
+    @property
+    def parsing_ctx(self) -> "TypeParsingCtx":
+        """A type parsing context derived from this checking context."""
+        from guppylang_internals.tys.parsing import TypeParsingCtx
+
+        return TypeParsingCtx(self.globals, self.generic_params)
+
 
 class DummyEvalDict(dict[str, Any]):
     """A custom dict that can be passed to `eval` to give better error messages.
