@@ -42,7 +42,9 @@ from guppylang_internals.checker.errors.type_errors import (
     MissingReturnValueError,
     StarredTupleUnpackError,
     TypeInferenceError,
+    TypeMismatchError,
     UnpackableError,
+    WrongNumberOfArgsError,
     WrongNumberOfUnpacksError,
 )
 from guppylang_internals.checker.expr_checker import (
@@ -58,6 +60,7 @@ from guppylang_internals.nodes import (
     DesugaredArrayComp,
     IterableUnpack,
     MakeIter,
+    ModifiedBlock,
     NestedFunctionDef,
     PlaceNode,
     TupleUnpack,
@@ -73,13 +76,15 @@ from guppylang_internals.tys.builtin import (
     is_sized_iter_type,
     nat_type,
 )
-from guppylang_internals.tys.const import ConstValue
+from guppylang_internals.tys.const import ConstValue, ExistentialConstVar
 from guppylang_internals.tys.parsing import type_from_ast
+from guppylang_internals.tys.qubit import is_qubit_ty, qubit_ty
 from guppylang_internals.tys.subst import Subst
 from guppylang_internals.tys.ty import (
     ExistentialTypeVar,
     FunctionType,
     NoneType,
+    NumericType,
     StructType,
     TupleType,
     Type,
@@ -356,7 +361,7 @@ class StmtChecker(AstVisitor[BBStatement]):
     def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.stmt:
         if node.value is None:
             raise GuppyError(UnsupportedError(node, "Variable declarations"))
-        ty = type_from_ast(node.annotation, self.ctx.globals, self.ctx.generic_params)
+        ty = type_from_ast(node.annotation, self.ctx.parsing_ctx)
         node.value, subst = self._check_expr(node.value, ty)
         assert not ty.unsolved_vars  # `ty` must be closed!
         assert len(subst) == 0
@@ -398,6 +403,48 @@ class StmtChecker(AstVisitor[BBStatement]):
         self.ctx.locals[func_def.name] = Variable(func_def.name, func_def.ty, func_def)
         return func_def
 
+    def visit_ModifiedBlock(self, node: ModifiedBlock) -> ast.stmt:
+        from guppylang_internals.checker.modifier_checker import check_modified_block
+
+        if not self.bb:
+            raise InternalGuppyError("BB required to check with block!")
+
+        # check the body of the modified block
+        modified_block = check_modified_block(node, self.bb, self.ctx)
+
+        # check the arguments of the control and power.
+        for control in modified_block.control:
+            ctrl = control.ctrl
+            # This case is handled during CFG construction.
+            assert len(ctrl) > 0
+            ctrl[0], ty = self._synth_expr(ctrl[0])
+
+            if is_array_type(ty):
+                if len(ctrl) > 1:
+                    span = Span(to_span(control.func).end, to_span(control).end)
+                    raise GuppyError(WrongNumberOfArgsError(span, 1, len(control.args)))
+                element_ty = get_element_type(ty)
+                if not is_qubit_ty(element_ty):
+                    n = ExistentialConstVar.fresh(
+                        "n", NumericType(NumericType.Kind.Nat)
+                    )
+                    dummy_array_ty = array_type(qubit_ty(), n)
+                    raise GuppyTypeError(TypeMismatchError(ctrl[0], dummy_array_ty, ty))
+                control.qubit_num = get_array_length(ty)
+            else:
+                for i in range(len(ctrl)):
+                    ctrl[i], subst = self._check_expr(ctrl[i], qubit_ty())
+                    assert len(subst) == 0
+                control.qubit_num = len(ctrl)
+
+        for power in node.power:
+            power.iter, subst = self._check_expr(
+                power.iter, NumericType(NumericType.Kind.Nat)
+            )
+            assert len(subst) == 0
+
+        return modified_block
+
     def visit_If(self, node: ast.If) -> None:
         raise InternalGuppyError("Control-flow statement should not be present here.")
 
